/*
 *	Copyright (C) 2009 Jonathan Naylor, G4KLX
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; version 2 of the License.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 */

#include "DStarReceiver.h"

#include <windows.h>
#include <tchar.h>

#include "WAVFileReader.h"
#include "DStarDefines.h"
#include "Utils.h"

#include <cassert>
#include <cctype>
#include <ctime>

const unsigned int FILENAME_LENGTH = 50U;

int _tmain(int argc, _TCHAR* argv[])
{
	if (argc < 2) {
		::_ftprintf(stderr, _T("Usage: D-Star Receiver <WAV File>\n"));
		return 1;
	}

	time_t now = ::time(NULL);
	struct tm* tm = ::localtime(&now);

	TCHAR fileName[FILENAME_LENGTH];
	::_tcsftime(fileName, FILENAME_LENGTH, _T("AMBE_%Y%m%d_%H%M%S.dat"), tm);

	FILE* fpAMBE = ::_tfopen(fileName, _T("wb"));
	if (fpAMBE == NULL) {
		::_ftprintf(stderr, _T("Cannot open the AMBE data file for writing - %s\n"), fileName);
		return 1;
	}

	::_ftprintf(stdout, _T("Writing AMBE data to %s\n"), fileName);

	::_tcsftime(fileName, FILENAME_LENGTH, _T("DATA_%Y%m%d_%H%M%S.dat"), tm);

	FILE* fpData = ::_tfopen(fileName, _T("wb"));
	if (fpData == NULL) {
		::_ftprintf(stderr, _T("Cannot open the DATA data file for writing - %s\n"), fileName);
		::fclose(fpAMBE);
		return 1;
	}

	::_ftprintf(stdout, _T("Writing [Slow] Data to %s\n"), fileName);

	::_tcsftime(fileName, FILENAME_LENGTH, _T("INFO_%Y%m%d_%H%M%S.dat"), tm);

	FILE* fpInfo = ::_tfopen(fileName, _T("wt"));
	if (fpInfo == NULL) {
		::_ftprintf(stderr, _T("Cannot open the INFO data file for writing - %s\n"), fileName);
		::fclose(fpAMBE);
		::fclose(fpData);
		return 1;
	}

	::_ftprintf(stdout, _T("Writing Info to %s\n"), fileName);

	CWAVFileReader reader(argv[1]);
	bool res = reader.open();
	if (!res) {
		::fclose(fpInfo);
		::fclose(fpData);
		::fclose(fpAMBE);
		return 1;
	}

	CDStarReceiver receiver(&reader, fpInfo, fpAMBE, fpData);
	receiver.run();

	reader.close();
	::fclose(fpInfo);
	::fclose(fpData);
	::fclose(fpAMBE);

	return 0;
}

const unsigned int BUFFER_LENGTH = 1600U * 8U;

CDStarReceiver::CDStarReceiver(ISoundReader* reader, FILE* fpInfo, FILE* fpAMBE, FILE* fpData) :
m_reader(reader),
m_modem(NULL),
m_header(),
m_frameMatcher(15U, FRAME_SYNC, 1U, 1000U),
m_dataMatcher(24U, DATA_SYNC, 1U, DATA_BLOCK_SIZE_BITS * 2U),
m_endMatcher(48U, END_PATTERN, 1U),
m_bitMatcher(4U, GMSK_BIT_SYNC, 16U),
m_scrambler(),
m_state(DSR_SEARCH_BIT_SYNC),
m_buffer(0),
m_length(0U),
m_count(21U),
m_dataLength(0U),
m_debug(false),
m_fpInfo(fpInfo),
m_fpAMBE(fpAMBE),
m_fpData(fpData)
{
	assert(reader != NULL);
	assert(fpInfo != NULL);
	assert(fpAMBE != NULL);
	assert(fpData != NULL);

	m_buffer = new bool[BUFFER_LENGTH];
}

CDStarReceiver::~CDStarReceiver()
{
	delete[] m_buffer;
}

void CDStarReceiver::run()
{
	unsigned int sampleRate = m_reader->getSampleRate();
	unsigned int channels   = m_reader->getChannels();

	float* buffer = new float[channels];

	m_modem = new CGMSKModem(sampleRate, GMSK_BIT_RATE, GMSK_BT);

	unsigned int length = m_reader->read(buffer, 1U);
	while (length > 0U) {
		MODEM_STATE state = m_modem->decode(buffer[0]);
	
		switch (state) {
			case MS_TRUE:
				addBit(true);
				break;
			case MS_FALSE:
				addBit(false);
				break;
			case MS_UNKNOWN:
				break;
		}

		length = m_reader->read(buffer, 1U);
	}

	delete m_modem;
	delete[] buffer;
}

void CDStarReceiver::addBit(bool bit)
{
	switch (m_state) {
		case DSR_SEARCH_BIT_SYNC: {
				bool res = m_bitMatcher.add(bit);
				if (res) {
					::fprintf(stdout, "Found bit sync\n");
					m_state = DSR_SEARCH_FRAME_SYNC;
					m_bitMatcher.reset();
				}
			}
			break;

		case DSR_SEARCH_FRAME_SYNC: {
				LPM_STATUS res = m_frameMatcher.add(bit);
				switch (res) {
					case LPM_FOUND:
						::fprintf(stdout, "Found frame sync\n");
						m_state = DSR_PROCESS_HEADER;
						m_frameMatcher.resetAll();
						m_length = 0U;
						break;

					case LPM_SEARCHING:
						break;

					case LPM_FAILED:
						::fprintf(stdout, "Cannot find frame sync\n");
						m_state = DSR_SEARCH_BIT_SYNC;
						m_frameMatcher.resetAll();
						break;
				}
			}
			break;

		case DSR_PROCESS_HEADER:
			m_buffer[m_length++] = bit;
			if (m_length == FEC_SECTION_LENGTH_BITS) {
				m_scrambler.reset();
				bool res = m_header.setRadioData(m_scrambler, m_buffer, m_length);
				if (!res) {
					m_length = 0U;
					m_state = DSR_SEARCH_BIT_SYNC;
					break;
				}

				interpretRadioHeader(); 

				if (m_header.isDataPacket()) {
					m_state = DSR_PROCESS_DD_DATA;
				} else {
					m_state = DSR_PROCESS_DV_VOICE;
					m_dataMatcher.resetAll();
					m_endMatcher.reset();
					m_count = 21U;
				}

				m_dataLength = 0U;
				m_length = 0U;
			}
			break;

		case DSR_PROCESS_DV_VOICE: {
				m_buffer[m_length++] = bit;
				bool res1       = m_endMatcher.add(bit);
				LPM_STATUS res2 = m_dataMatcher.add(bit);

				if (res1) {
					::fprintf(stdout, "Found end pattern\n\n");
					m_state = DSR_SEARCH_BIT_SYNC;
					m_length = 0U;
					break;
				} else if (m_length == DV_FRAME_LENGTH_BITS) {
					if (m_count == 21U) {
						if (res2 == LPM_FOUND) {
							processDVVoice(m_buffer);
							::fprintf(stdout, "Re-synchronised the data\n");
							m_dataMatcher.resetAll();
							m_count = 0U;
						} else {
							::fprintf(stdout, "Cannot find the data re-sync vector\n");
							m_state = DSR_SEARCH_BIT_SYNC; // DSR_SEARCH_DATA_SYNC;
							m_dataMatcher.resetCount();
						}
					} else {
						processDVVoice(m_buffer);
						processDVSlowData(m_buffer + VOICE_FRAME_LENGTH_BITS);
					}

					m_count++;
					m_length = 0U;
				}
			}
			break;

		// The data sync has been missed, search for it
		case DSR_SEARCH_DATA_SYNC: {
				m_buffer[m_length++] = bit;
				LPM_STATUS res = m_dataMatcher.add(bit);

				switch (res) {
					case LPM_FOUND:
						::fprintf(stdout, "Re-synchronised the data\n");
						processDVVoice(m_buffer + m_length - DV_FRAME_LENGTH_BITS);
						m_state = DSR_PROCESS_DV_VOICE;
						m_dataMatcher.resetAll();
						m_length = 0U;
						m_count  = 0U;
						break;

					case LPM_SEARCHING:
						break;

					case LPM_FAILED:
						::fprintf(stdout, "Timed-out when waiting for a data re-sync vector\n");
						m_state = DSR_SEARCH_BIT_SYNC;
						m_dataMatcher.resetAll();
						m_length = 0U;
						break;
				}
			}
			break;

		case DSR_PROCESS_DD_DATA: {
				m_buffer[m_length++] = bit;

				if (m_length == 16U) {
					// Descramble the length field
					bool buffer[16U];
					m_scrambler.process(m_buffer, buffer, 16U);

					m_dataLength  = CUtils::bitsToByte(buffer + 8U) << 8U;
					m_dataLength |= CUtils::bitsToByte(buffer + 0U) << 0U;
				} else if (m_length == (20U + m_dataLength) * 8U) {
					bool res = processDDData();
					if (!res)
						::fprintf(stderr, "Checksum failed on the data\n");

					m_state = DSR_SEARCH_BIT_SYNC;
					m_length = 0U;
				}
			}
			break;
	}
}

void CDStarReceiver::debug(bool set)
{
	m_debug = set;

	if (m_modem != NULL)
		m_modem->debug(set);
	m_header.debug(set);
	m_bitMatcher.debug(set);
	m_endMatcher.debug(set);
	m_dataMatcher.debug(set);
	m_frameMatcher.debug(set);
}

void CDStarReceiver::interpretRadioHeader()
{
	if (m_header.isDataPacket())
		::fprintf(m_fpInfo, "Type: Data packet\n");
	else
		::fprintf(m_fpInfo, "Type: Voice packet\n");

	if (m_header.isDirectMode())
		::fprintf(m_fpInfo, "Mode: Direct connection\n");
	else
		::fprintf(m_fpInfo, "Mode: Via repeater\n");

	if (m_header.isInterrupted())
		::fprintf(m_fpInfo, "Interrupted: Yes\n");
	else
		::fprintf(m_fpInfo, "Interrupted: No\n");

	if (m_header.isControlSignal())
		::fprintf(m_fpInfo, "Control signal: Yes\n");
	else
		::fprintf(m_fpInfo, "Control signal: No\n");

	if (m_header.isUrgent())
		::fprintf(m_fpInfo, "Urgent priority: Yes\n");
	else
		::fprintf(m_fpInfo, "Urgent priority: No\n");

	::fprintf(m_fpInfo, "Repeater flags: ");

	switch (m_header.getRepeaterFlags()) {
		case DSRH_REPEATER_CONTROL:
			::fprintf(m_fpInfo, "Control mode\n");
			break;
		case DSRH_AUTO_REPLY:
			::fprintf(m_fpInfo, "Auto reply\n");
			break;
		case DSRH_REQUEST_RESEND:
			::fprintf(m_fpInfo, "Request resend\n");
			break;
		case DSRH_ACK_FLAG:
			::fprintf(m_fpInfo, "ACK flag\n");
			break;
		case DSRH_NO_RESPONSE:
			::fprintf(m_fpInfo, "No response available\n");
			break;
		case DSRH_RELAY_UNAVAILABLE:
			::fprintf(m_fpInfo, "Relay unavailable\n");
			break;
		default:
			::fprintf(m_fpInfo, "None\n");
			break;
	}

	::fprintf(m_fpInfo, "Destination repeater callsign: %s\n", m_header.getDestinationRepeaterCallsign());
	::fprintf(m_fpInfo, "Departure repeater callsign: %s\n", m_header.getDepartureRepeaterCallsign());
	::fprintf(m_fpInfo, "Companion callsign: %s\n", m_header.getCompanionCallsign());
	::fprintf(m_fpInfo, "Own callsign 1: %s\n", m_header.getOwnCallsign1());
	::fprintf(m_fpInfo, "Own callsign 2: %s\n", m_header.getOwnCallsign2());
}

void CDStarReceiver::processDVVoice(const bool* data)
{
	unsigned char buffer[VOICE_FRAME_LENGTH_BYTES];

	unsigned int n = 0U;
	for (unsigned int i = 0U; i < VOICE_FRAME_LENGTH_BYTES; i++, n += 8U)
		buffer[i] = CUtils::bitsToByteRev(data + n);

	::fwrite(buffer, 1U, VOICE_FRAME_LENGTH_BYTES, m_fpAMBE);
}

void CDStarReceiver::processDVSlowData(const bool* data)
{
	m_scrambler.reset();

	bool buffer1[DATA_FRAME_LENGTH_BITS];
	m_scrambler.process(data, buffer1, DATA_FRAME_LENGTH_BITS);

	unsigned char buffer2[DATA_FRAME_LENGTH_BYTES];
	unsigned int n = 0U;
	for (unsigned int i = 0U; i < DATA_FRAME_LENGTH_BYTES; i++, n += 8U)
		buffer2[i] = CUtils::bitsToByteRev(buffer1 + n);

	::fwrite(buffer2, 1U, DATA_FRAME_LENGTH_BYTES, m_fpData);
}

bool CDStarReceiver::processDDData()
{
	bool* buffer1 = new bool[m_length];

	m_scrambler.process(m_buffer, buffer1, m_length);

	unsigned int bytes = m_length / 8U;

	unsigned char* buffer2 = new unsigned char[bytes];

	unsigned int n = 0U;
	for (unsigned int i = 0U; i < bytes; i++, n += 8U)
		buffer2[i] = CUtils::bitsToByteRev(buffer1 + n);

	::fwrite(buffer2, 1U, bytes, m_fpData);

	delete[] buffer1;
	delete[] buffer2;

	return true;
}
